﻿using System;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// A base class for operators which evaluate TSP solutions.
  /// </summary>
  [Item("MetaOptimizationEvaluator", "A base class for operators which evaluate Meta Optimization solutions.")]
  [StorableClass]
  public class MetaOptimizationEvaluator : SingleSuccessorOperator, IMetaOptimizationEvaluator {
    private const string RepetitionsParameterName = "Repetitions";

    private bool algorithmStopped;

    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<EngineAlgorithm> AlgorithmParameter {
      get { return (ILookupParameter<EngineAlgorithm>)Parameters["Algorithm"]; }
    }
    public ILookupParameter<IItemList<ISingleObjectiveProblem>> ProblemsParameter {
      get { return (ILookupParameter<IItemList<ISingleObjectiveProblem>>)Parameters["Problems"]; }
    }
    public ILookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ILookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ValueParameter<IntValue> RepetitionsParameter {
      get { return (ValueParameter<IntValue>)Parameters[RepetitionsParameterName]; }
    }

    public IntValue Repetitions {
      get { return RepetitionsParameter.Value; }
      set { RepetitionsParameter.Value = value; }
    }

    public MetaOptimizationEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The evaluated quality of the ParameterVector."));
      Parameters.Add(new LookupParameter<EngineAlgorithm>("Algorithm", "Missing description."));
      Parameters.Add(new LookupParameter<IItemList<ISingleObjectiveProblem>>("Problems", "Missing description."));
      Parameters.Add(new LookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", "Missing description."));
      Parameters.Add(new ValueParameter<IntValue>(RepetitionsParameterName, "Number of evaluations for one individual.", new IntValue(3)));
    }

    [StorableConstructor]
    protected MetaOptimizationEvaluator(bool deserializing) : base(deserializing) { }
    protected MetaOptimizationEvaluator(MetaOptimizationEvaluator original, Cloner cloner)
      : base(original, cloner) {
      this.algorithmStopped = original.algorithmStopped;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new MetaOptimizationEvaluator(this, cloner);
    }

    public override IOperation Apply() {
      EngineAlgorithm algorithm = AlgorithmParameter.ActualValue;
      IItemList<ISingleObjectiveProblem> problems = ProblemsParameter.ActualValue;

      // set parameters
      ParameterConfigurationParameter.ActualValue.Parameterize(algorithm);
      algorithm.Problem = problems.First();
      algorithm.StoreAlgorithmInEachRun = false;

      algorithmStopped = false;
      algorithm.Stopped += new EventHandler(ActualValue_Stopped);

      List<double> qualities = new List<double>();
      List<TimeSpan> executionTimes = new List<TimeSpan>();
      algorithm.Engine = new SequentialEngine.SequentialEngine();
      algorithm.Prepare(true);

      foreach (ISingleObjectiveProblem problem in problems) {
        algorithm.Problem = problem;

        for (int i = 0; i < Repetitions.Value; i++) {
          algorithm.Prepare();
          algorithm.Start();
          while (!algorithmStopped) {
            Thread.Sleep(200); // wait for algorithm to complete; do not rely on Algorithm.ExecutionState here, because change of ExecutionState happens before Run is added (which causes problems because Algorithm might get cloned when its started already)
          }
          qualities.Add(((DoubleValue)algorithm.Results["BestQuality"].Value).Value);
          executionTimes.Add(algorithm.ExecutionTime);

          // parameters will be stored in ParameterConfigurationTree anyway. they would be redundant in runs
          algorithm.Runs.Last().Parameters.Clear();
          // but keep the problem, since this differs in runs
          algorithm.Runs.Last().Parameters.Add("Problem", problem);
          algorithmStopped = false;
        }

      }


      algorithm.Stopped -= new EventHandler(ActualValue_Stopped);
      algorithm.Prepare();

      qualities = qualities.OrderBy(x => x).ToList();  // todo: respect Maximization:true/false

      ParameterConfigurationParameter.ActualValue.AverageExecutionTime = new TimeSpanValue(TimeSpan.FromMilliseconds(executionTimes.Average(t => t.TotalMilliseconds)));
      ParameterConfigurationParameter.ActualValue.Repetitions = Repetitions;
      ParameterConfigurationParameter.ActualValue.BestQuality = new DoubleValue(qualities.First());
      ParameterConfigurationParameter.ActualValue.AverageQuality = new DoubleValue(qualities.Average());
      ParameterConfigurationParameter.ActualValue.WorstQuality = new DoubleValue(qualities.Last());
      ParameterConfigurationParameter.ActualValue.QualityVariance = new DoubleValue(qualities.Variance());
      ParameterConfigurationParameter.ActualValue.QualityStandardDeviation = new DoubleValue(qualities.StandardDeviation());
      ParameterConfigurationParameter.ActualValue.Runs = algorithm.Runs;

      double quality = ParameterConfigurationParameter.ActualValue.AverageQuality.Value; // todo: also include other measures (executiontime, variance)
      this.QualityParameter.ActualValue = new DoubleValue(quality);

      return base.Apply();
    }

    public static double Variance(IEnumerable<double> source) {
      double avg = source.Average();
      double d = source.Aggregate(0.0, (total, next) => total += Math.Pow(next - avg, 2));
      return d / (source.Count() - 1);
    }

    public static double StandardDeviation(IEnumerable<double> source) {
      return Math.Sqrt(source.Variance());
    }

    void ActualValue_Stopped(object sender, EventArgs e) {
      algorithmStopped = true;
    }
  }
}
